@@ -1,5 +1,5 @@
 #
-#   engn/perldb2/DB2.pm, engn_perldb2, db2_v81, 1.6 00/09/06 15:57:59
+#   engn/perldb2/DB2.pm, engn_perldb2, db2_v81, 1.7 01/03/29 17:06:45
 #
 #   Copyright (c) 1995,1996,1997,1998,1999,2000  International Business Machines Corp.
 #
@@ -23,7 +23,7 @@
                      $attrib_clobfile
                      $attrib_dbclobfile );
 
-    $VERSION = '0.74';
+    $VERSION = '0.75';
         require_version DBI 0.93;
 
     bootstrap DBD::DB2;
@@ -195,6 +195,11 @@
         $sth;
     }
 
+    sub ping {
+        my($dbh) = @_;
+        DBD::DB2::db::_ping( $dbh );
+    }
+
     sub tables {
         my ($dbh) = @_;
         my $tablesref = DBD::DB2::db::_tables( $dbh );
