@@ -1,5 +1,5 @@
 /*
-   dbdimp.c, engn_perldb2, db2_v81, 1.6 00/09/06 17:18:26
+   engn/perldb2/dbdimp.c, engn_perldb2, db2_v81, 1.7 01/03/29 17:06:48
 
    Copyright (c) 1995,1996,1997,1998,1999,2000 International Business Machines Corp.
 */
@@ -18,6 +18,14 @@
  #define SQL_ATTR_DB2_SQLERRP  2451
 #endif
 
+#ifndef SQL_ATTR_OPTIMIZE_FOR_NROWS
+ #define SQL_ATTR_OPTIMIZE_FOR_NROWS 2450
+#endif
+
+#ifndef SQL_ATTR_QUERY_OPTIMIZATION_LEVEL
+ #define SQL_ATTR_QUERY_OPTIMIZATION_LEVEL 1293
+#endif
+
 DBISTATE_DECLARE;
 
 void dbd_init( dbistate_t *dbistate )
@@ -39,9 +47,15 @@ static SQLRETURN do_error( SV *h,
     SQLCHAR msgBuffer[SQL_MAX_MESSAGE_LENGTH+1];
     char *msg = NULL;
 
-    if( SQL_SUCCESS == rc || SQL_NO_DATA == rc )
+    if( SQL_SUCCESS == rc )
       return rc;
 
+    if( SQL_NO_DATA == rc )
+    {
+      sv_setpv( state, SQL_NODATA_EXCEPTION );
+      return rc;
+    }
+
     if( SQL_SUCCESS_WITH_INFO == rc && DBIS->debug < 3 )
       return SQL_SUCCESS;
 
@@ -91,9 +105,15 @@ static SQLRETURN check_error( SV *h,
     SQLSMALLINT handleType;
     SQLHANDLE handle = SQL_NULL_HANDLE;
 
-    if( SQL_SUCCESS == rc || SQL_NO_DATA == rc )
+    if( SQL_SUCCESS == rc )
       return rc;
 
+    if( SQL_NO_DATA == rc )
+    {
+      sv_setpv( DBIc_STATE( imp_xxh ), SQL_NODATA_EXCEPTION );
+      return rc;
+    }
+
     if( SQL_SUCCESS_WITH_INFO == rc && DBIS->debug < 3 )
       return SQL_SUCCESS;
 
@@ -394,6 +414,104 @@ int dbd_db_do( SV *dbh,
 }
 
 
+int dbd_db_ping( SV *dbh )
+{
+    D_imp_dbh(dbh);
+    SQLRETURN ret;
+    SQLHSTMT stmt = SQL_NULL_HSTMT;
+    const char *pSQL = "values 1";
+
+    if( !DBIc_ACTIVE( imp_dbh ) )
+      return 0;
+
+    if( '\0' == imp_dbh->sqlerrp[0] )
+    {
+      SQLGetConnectAttr( imp_dbh->hdbc,
+                         SQL_ATTR_DB2_SQLERRP,
+                         (SQLPOINTER)imp_dbh->sqlerrp,
+                         sizeof( imp_dbh->sqlerrp ),
+                         NULL );
+    }
+
+    if( strncmp( imp_dbh->sqlerrp, "SQL", 3 ) == 0 )      /* UNO */
+    {
+      /* Do nothing, use the default statement */
+    }
+    else if( strncmp( imp_dbh->sqlerrp, "DSN", 3 ) == 0 ) /* MVS */
+    {
+      pSQL = "select 1 from sysibm.sysdummy1";
+    }
+    else if( strncmp( imp_dbh->sqlerrp, "QSQ", 3 ) == 0 ) /* AS/400 */
+    {
+      pSQL = "select 1 from qsys2.qsqptabl";
+    }
+    else if( strncmp( imp_dbh->sqlerrp, "ARI", 3 ) == 0 ) /* VM */
+    {
+      pSQL = "select 1 from system.sysoptions";
+    }
+    else
+    {
+      /* Do nothing, use the default statement */
+    }
+
+    ret = SQLAllocHandle( SQL_HANDLE_STMT, imp_dbh->hdbc, &stmt );
+    ret = do_error( dbh,
+                    ret,
+                    stmt == SQL_NULL_HSTMT ? SQL_HANDLE_DBC
+                                           : SQL_HANDLE_STMT,
+                    stmt == SQL_NULL_HSTMT ? imp_dbh->hdbc
+                                           : stmt,
+                    "dbd_db_ping: Statement allocation error" );
+    if( SQL_SUCCESS != ret )
+       goto exit;
+
+    ret = SQLSetStmtAttr( stmt,
+                          SQL_ATTR_DEFERRED_PREPARE,
+                          SQL_DEFERRED_PREPARE_OFF,
+                          0 );
+    ret = do_error( dbh,
+                    ret,
+                    SQL_HANDLE_STMT,
+                    stmt,
+                    "dbd_db_ping: Error turning off deferred prepare" );
+    if( SQL_SUCCESS != ret )
+       goto exit;
+
+    ret = SQLPrepare( stmt, (SQLCHAR*)pSQL, SQL_NTS );
+    ret = do_error( dbh,
+                    ret,
+                    SQL_HANDLE_STMT,
+                    stmt,
+                    "dbd_db_ping: Error preparing statement" );
+
+exit:
+    if( stmt )
+      SQLFreeHandle( SQL_HANDLE_STMT, stmt );
+
+    /* If any error occured, check the state to determine if the cause */
+    /* is a broken connection.                                         */
+    if( SQL_SUCCESS != ret )
+    {
+      STRLEN len;
+      char *pState = SvPV( DBIc_STATE(imp_dbh), len );
+
+      if( pState && len >= 5 &&
+          ( strncmp( pState, "08", 2 ) == 0 ||
+            strncmp( pState, "40003", 5 ) == 0 ) )
+      {
+        /* ping should not throw an error when it detects a dead  */
+        /* connection so reset error code and message but keep    */
+        /* connection state                                       */
+        sv_setsv( DBIc_ERRSTR(imp_dbh), &sv_undef );
+        sv_setsv( DBIc_ERR(imp_dbh), &sv_undef );
+        return 0; /* Connection is dead */
+      }
+    }
+
+    return 1; /* Connection is still alive */
+}
+
+
 int dbd_db_commit( SV *dbh,
                    imp_dbh_t *imp_dbh )
 {
@@ -466,35 +584,49 @@ void dbd_db_destroy( SV *dbh,
 static SQLINTEGER getConnectAttr( char *key,
                                   STRLEN keylen )
 {
+  /*
+   * The following DB2 CLI connection attributes are not supported
+   *
+   *   SQL_ATTR_ASYNC_ENABLE        Doesn't make sense for DBD::DB2
+   *   SQL_ATTR_AUTO_IPD            Doesn't make sense for DBD::DB2
+   *   SQL_ATTR_CONNECTION_DEAD     not reliable, $dbh->ping is better
+   *   SQL_ATTR_CONNECTTYPE         2-phase commit not supported
+   *   SQL_ATTR_CONN_CONTEXT        Doesn't make sense for DBD::DB2
+   *   SQL_ATTR_ENLIST_IN_DTC       Doesn't make sense for DBD::DB2
+   *   SQL_ATTR_LOGIN_TIMEOUT       Not supported by DB2 CLI
+   *   SQL_ATTR_MAXCONN             For NetBIOS
+   *   SQL_ATTR_OPTIMIZE_SQLCOLUMNS
+   *   SQL_ATTR_SYNC_POINT          2-phase commit not supported
+   *   SQL_ATTR_TRANSLATE_LIB       Not supported by DB2 CLI
+   *   SQL_ATTR_TRANSLATE_OPTION    Not supported by DB2 CLI
+   *   SQL_ATTR_WCHARTYPE           Doesn't make sense for DBD::DB2
+   *
+   */
+
   /* For better performance, the keys are sorted by length */
   switch( keylen )
   {
     case 10:
       if(      strEQ( key, "AutoCommit" ) )
         return SQL_ATTR_AUTOCOMMIT;
-      return 0;
+      return -1;
 
     case 11:
       if(      strEQ( key, "db2_sqlerrp" ) )
         return SQL_ATTR_DB2_SQLERRP;
-      return 0;
-
-    case 12:
-      if(      strEQ( key, "db2_auto_ipd" ) )
-        return SQL_ATTR_AUTO_IPD;
-      return 0;
+      return -1;
 
     case 13:
       if(      strEQ( key, "db2_clischema" ) )
         return SQL_ATTR_CLISCHEMA;
-      return 0;
+      return -1;
 
     case 14:
       if(      strEQ( key, "db2_db2explain" ) )
         return SQL_ATTR_DB2EXPLAIN;
       else if( strEQ( key, "db2_quiet_mode" ) )
         return SQL_ATTR_QUIET_MODE;
-      return 0;
+      return -1;
 
     case 15:
       if(      strEQ( key, "db2_access_mode" ) )
@@ -503,40 +635,38 @@ static SQLINTEGER getConnectAttr( char *key,
         return SQL_ATTR_DB2ESTIMATE;
       else if( strEQ( key, "db2_info_userid" ) )
         return SQL_ATTR_INFO_USERID;
-      return 0;
+      return -1;
 
     case 16:
-      if(      strEQ( key, "db2_async_enable" ) )
-        return SQL_ATTR_ASYNC_ENABLE;
-      else if( strEQ( key, "db2_connect_node" ) )
+      if(      strEQ( key, "db2_connect_node" ) )
         return SQL_ATTR_CONNECT_NODE;
       else if( strEQ( key, "db2_info_acctstr" ) )
         return SQL_ATTR_INFO_ACCTSTR;
-      return 0;
+      return -1;
 
     case 17:
       if(      strEQ( key, "db2_info_applname" ) )
         return SQL_ATTR_INFO_APPLNAME;
       else if( strEQ( key, "db2_txn_isolation" ) )
         return SQL_ATTR_TXN_ISOLATION;
-      return 0;
+      return -1;
 
     case 18:
       if(      strEQ( key, "db2_close_behavior" ) )
         return SQL_ATTR_CLOSE_BEHAVIOR;
       else if( strEQ( key, "db2_current_schema" ) )
         return SQL_ATTR_CURRENT_SCHEMA;
-      return 0;
+      return -1;
 
     case 19:
-      if(      strEQ( key, "db2_info_wrkstnname" ) )
+      if( strEQ( key, "db2_info_wrkstnname" ) )
         return SQL_ATTR_INFO_WRKSTNNAME;
       else if( strEQ( key, "db2_longdata_compat" ) )
         return SQL_ATTR_LONGDATA_COMPAT;
-      return 0;
+      return -1;
 
     default:
-      return 0;
+      return -1;
   }
 }
 
@@ -554,29 +684,12 @@ int dbd_db_STORE_attrib( SV *dbh,
   SQLINTEGER StringLength = 0;
   char msg[128]; /* buffer for error messages */
 
-  if( 0 == Attribute ) /* Don't know what this attribute is */
+  if( Attribute < 0 ) /* Don't know what this attribute is */
     return FALSE;
 
-  /*
-   * The following DB2 CLI connection attributes are not supported
-   *
-   *   SQL_ATTR_AUTO_IPD            Read only
-   *   SQL_ATTR_CONNECTTYPE         2-phase commit not supported
-   *   SQL_ATTR_CONN_CONTEXT        Doesn't make sense for DBD::DB2
-   *   SQL_ATTR_DB2_SQLERRP         Read only
-   *   SQL_ATTR_ENLIST_IN_DTC       Doesn't make sense for DBD::DB2
-   *   SQL_ATTR_LOGIN_TIMEOUT       Not supported by DB2 CLI
-   *   SQL_ATTR_MAXCONN             For NetBIOS
-   *   SQL_ATTR_OPTIMIZE_SQLCOLUMNS
-   *   SQL_ATTR_SYNC_POINT          2-phase commit not supported
-   *   SQL_ATTR_TRANSLATE_OPTION    Not supported by DB2 CLI
-   *   SQL_ATTR_WCHARTYPE           Doesn't make sense for DBD::DB2
-   *
-   */
   switch( Attribute )
   {
     /* Booleans */
-    case SQL_ATTR_ASYNC_ENABLE:
     case SQL_ATTR_AUTOCOMMIT:
     case SQL_ATTR_LONGDATA_COMPAT:
       if( SvTRUE( valuesv ) )
@@ -665,7 +778,7 @@ SV *dbd_db_FETCH_attrib( SV *dbh,
   SQLINTEGER BufferLength = sizeof( buffer );
   SQLINTEGER StringLength;
 
-  if( 0 == Attribute ) /* Not a DB2 attribute */
+  if( Attribute < 0 ) /* Don't know what this attribute is */
     return NULL;
 
   ret = SQLGetConnectAttr( imp_dbh->hdbc,
@@ -692,9 +805,7 @@ SV *dbd_db_FETCH_attrib( SV *dbh,
     switch( Attribute )
     {
       /* Booleans */
-      case SQL_ATTR_ASYNC_ENABLE:
       case SQL_ATTR_AUTOCOMMIT:
-      case SQL_ATTR_AUTO_IPD:
       case SQL_ATTR_LONGDATA_COMPAT:
         if( *(SQLINTEGER*)ValuePtr )
           retsv = &sv_yes;
@@ -876,11 +987,20 @@ static int dbd_describe( SV *sth,
 
     if (imp_sth->done_desc)
         return 1;    /* success, already done it */
-    imp_sth->done_desc = 1;
 
     ret = SQLNumResultCols(imp_sth->phstmt,&num_fields);
     ret = check_error( sth, ret, "SQLNumResultCols failed" );
     EOI(ret);
+
+    /* If execute hasn't been called yet and num_fields is zero it */
+    /* might mean that this is a CALL statement in which case we   */
+    /* must wait until after the execute to describe.  Just return */
+    /* without setting done_desc flag.                             */
+    if( 0 == num_fields && !DBIc_ACTIVE( imp_sth ) )
+      return 1;
+
+    imp_sth->done_desc = 1;
+
     DBIc_NUM_FIELDS(imp_sth) = num_fields;
 
     if( 0 == num_fields )
@@ -1137,10 +1257,10 @@ int dbd_st_prepare( SV *sth,
     if (params > 0 ){
         /* scan statement for '?', ':1' and/or ':foo' style placeholders*/
         dbd_preparse(imp_sth, statement);
-    } else {    /* assuming a parameterless select */
-        dbd_describe(sth,imp_sth );
     }
 
+    dbd_describe( sth, imp_sth );
+
     /* initialize sth pointers */
     imp_sth->RowCount = -1;
     imp_sth->bHasInput = 0;
@@ -1489,17 +1609,17 @@ int dbd_st_execute( SV *sth,     /* error : <=(-2), ok row count : >=0, unknown
       }
     }
 
+    ret = SQLExecute(imp_sth->phstmt);
+    ret = check_error( sth, ret, "SQLExecute failed" );
+    if (ret < 0)
+        return(-2);
+
     /* describe and allocate storage for results        */
     if (!imp_sth->done_desc && !dbd_describe(sth, imp_sth)) {
         /* dbd_describe has already called check_error()        */
         return -2;
     }
 
-    ret = SQLExecute(imp_sth->phstmt);
-    ret = check_error( sth, ret, "SQLExecute failed" );
-    if (ret < 0)
-        return(-2);
-
     if( imp_sth->bind_names && imp_sth->bHasOutput )
     {
       hv = imp_sth->bind_names;
@@ -1568,8 +1688,8 @@ AV *dbd_st_fetch( SV *sth,
             if (DBIS->debug >= 3)
                 fprintf( (FILE*)DBILOGFP,
                          "    dbd_st_fetch failed, rc=%d", ret);
-            check_error( sth, ret, "Fetch failed" );
         }
+        check_error( sth, ret, "Fetch failed" );
         dbd_st_finish(sth, imp_sth);
         return Nullav;
     }
@@ -1784,13 +1904,142 @@ void dbd_st_destroy( SV *sth,
 }
 
 
+static SQLINTEGER getStatementAttr( char *key,
+                                    STRLEN keylen )
+{
+  /* For better performance, the keys are sorted by length */
+  switch( keylen )
+  {
+    case 10:
+      if(      strEQ( key, "db2_noscan" ) )
+        return SQL_ATTR_NOSCAN;
+      return -1;
+
+    case 12:
+      if(      strEQ( key, "db2_max_rows" ) )
+        return SQL_ATTR_MAX_ROWS;
+      else if( strEQ( key, "db2_prefetch" ) )
+        return SQL_ATTR_PREFETCH;
+      return -1;
+
+    case 14:
+      if(      strEQ( key, "db2_earlyclose" ) )
+        return SQL_ATTR_EARLYCLOSE;
+      else if( strEQ( key, "db2_max_length" ) )
+        return SQL_ATTR_MAX_LENGTH;
+      else if( strEQ( key, "db2_row_number" ) )
+        return SQL_ATTR_ROW_NUMBER;
+      return -1;
+
+    case 15:
+      if(      strEQ( key, "db2_concurrency" ) )
+        return SQL_ATTR_CONCURRENCY;
+      else if( strEQ( key, "db2_cursor_hold" ) )
+        return SQL_ATTR_CURSOR_HOLD;
+      return -1;
+
+    case 17:
+      if(      strEQ( key, "db2_query_timeout" ) )
+        return SQL_ATTR_QUERY_TIMEOUT;
+      else if( strEQ( key, "db2_retrieve_data" ) )
+        return SQL_ATTR_RETRIEVE_DATA;
+      else if( strEQ( key, "db2_txn_isolation" ) )
+        return SQL_ATTR_TXN_ISOLATION;
+      return -1;
+
+    case 20:
+      if(      strEQ( key, "db2_deferred_prepare" ) )
+        return SQL_ATTR_DEFERRED_PREPARE;
+      return -1;
+
+    case 22:
+      if(      strEQ( key, "db2_optimize_for_nrows" ) )
+        return SQL_ATTR_OPTIMIZE_FOR_NROWS;
+      return -1;
+
+    case 28:
+      if(      strEQ( key, "db2_query_optimization_level" ) )
+        return SQL_ATTR_QUERY_OPTIMIZATION_LEVEL;
+      return -1;
+
+    default:
+      return -1;
+  }
+}
+
+
 int dbd_st_STORE_attrib( SV *sth,
                          imp_sth_t *imp_sth,
                          SV *keysv,
                          SV *valuesv )
 {
-  /* No (writeable) DB2 statement attributes supported yet */
-  return FALSE;
+  STRLEN kl;
+  char *key = SvPV( keysv, kl );
+  SQLINTEGER Attribute = getStatementAttr( key, kl );
+  SQLRETURN ret;
+  SQLPOINTER ValuePtr = 0;
+  SQLINTEGER StringLength = 0;
+  char msg[128]; /* buffer for error messages */
+
+  if( Attribute < 0 ) /* Don't know what this attribute is */
+    return FALSE;
+
+  switch( Attribute )
+  {
+    /* Booleans */
+    case SQL_ATTR_CURSOR_HOLD:
+    case SQL_ATTR_DEFERRED_PREPARE:
+    case SQL_ATTR_EARLYCLOSE:
+    case SQL_ATTR_NOSCAN:
+    case SQL_ATTR_PREFETCH:
+    case SQL_ATTR_RETRIEVE_DATA:
+      if( SvTRUE( valuesv ) )
+        ValuePtr = (SQLPOINTER)1;
+      break;
+
+    /* Integers */
+    case SQL_ATTR_CONCURRENCY:
+    case SQL_ATTR_MAX_LENGTH:
+    case SQL_ATTR_MAX_ROWS:
+    case SQL_ATTR_OPTIMIZE_FOR_NROWS:
+    case SQL_ATTR_QUERY_OPTIMIZATION_LEVEL:
+    case SQL_ATTR_QUERY_TIMEOUT:
+    case SQL_ATTR_TXN_ISOLATION:
+      if( SvIOK( valuesv ) )
+      {
+        ValuePtr = (SQLPOINTER)SvIV( valuesv );
+      }
+      else if( SvOK( valuesv ) )
+      {
+        /* Value is not an integer, return error */
+        sprintf( msg,
+                 "Invalid value for statement attribute %s, expecting integer",
+                 key );
+        do_error( sth, -1, 0, SQL_NULL_HANDLE, msg );
+        return FALSE;
+      }
+      else /* Undefined, Set to default, most are 0 or NULL */
+      {
+      }
+      break;
+
+    default:
+      return FALSE;
+  }
+
+
+  ret = SQLSetStmtAttr( imp_sth->phstmt,
+                        Attribute,
+                        ValuePtr,
+                        StringLength );
+  if( SQL_SUCCESS != ret )
+  {
+    sprintf( msg, "Error setting %s statement attribute", key );
+    check_error( sth, ret, msg );
+    return FALSE;
+  }
+
+  return TRUE;
 }
 
 
@@ -1803,84 +2052,166 @@ SV *dbd_st_FETCH_attrib( SV *sth,
   int i;
   SV *retsv = NULL;
   AV *av;
+  int cacheit = 1;
+  SQLINTEGER Attribute;
 
-  if (!imp_sth->done_desc && !dbd_describe(sth, imp_sth)) {
-      /* dbd_describe has already called check_error()        */
-      /* We can't return Nullsv here because the xs code will */
-      /* then just pass the attribute name to DBI for FETCH.  */
-      croak("Describe failed during %s->FETCH(%s)",
-             SvPV(sth,na), key);
+  if (!imp_sth->done_desc && !dbd_describe(sth, imp_sth))
+  {
+    /* dbd_describe has already called check_error()        */
+    /* We can't return Nullsv here because the xs code will */
+    /* then just pass the attribute name to DBI for FETCH.  */
+    croak("Describe failed during %s->FETCH(%s)",
+           SvPV(sth,na), key);
   }
 
   i = DBIc_NUM_FIELDS(imp_sth);
 
-  if (kl==7 && strEQ(key, "lengths")) {
-      av = newAV();
-      retsv = newRV((SV*)av);
-      while(--i >= 0)
-          av_store(av, i, newSViv((IV)imp_sth->fbh[i].dsize));
-  } else if (kl==5 && strEQ(key, "types")) {
-      av = newAV();
-      retsv = newRV((SV*)av);
-      while(--i >= 0)
-          av_store(av, i, newSViv(imp_sth->fbh[i].dbtype));
+  if (kl==7 && strEQ(key, "lengths"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( (SV*)av ) );
+    while(--i >= 0)
+        av_store(av, i, newSViv((IV)imp_sth->fbh[i].dsize));
+  }
+  else if (kl==5 && strEQ(key, "types"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( (SV*)av ) );
+    while(--i >= 0)
+        av_store(av, i, newSViv(imp_sth->fbh[i].dbtype));
+  }
+  else if (kl==13 && strEQ(key, "NUM_OF_PARAMS"))
+  {
+    HV *bn = imp_sth->bind_names;
+    retsv = sv_2mortal( newSViv( (bn) ? HvKEYS(bn) : 0 ) );
+  }
+  else if (kl==4 && strEQ(key, "NAME"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( (SV*)av ) );
+    while(--i >= 0)
+        av_store(av, i, newSVpv((char *)imp_sth->fbh[i].cbuf,0));
+  }
+  else if (kl==8 && strEQ(key, "NULLABLE"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( (SV*)av ) );
+    while(--i >= 0)
+        av_store(av, i,
+                 (imp_sth->fbh[i].nullok == 1) ? &sv_yes : &sv_no);
+  }
+  else if (kl==10 && strEQ(key, "CursorName"))
+  {
+    char cursor_name[256];
+    SQLSMALLINT cursor_name_len;
+    SQLRETURN ret =
+          SQLGetCursorName(imp_sth->phstmt, (SQLCHAR *)cursor_name,
+                           sizeof(cursor_name), &cursor_name_len);
+    ret = check_error( sth, ret, "SQLGetCursorName failed" );
+    if (ret < 0)
+        return Nullsv;
+    else
+        retsv = sv_2mortal( newSVpv(cursor_name, cursor_name_len) );
+  }
+  else if (kl==4 && strEQ(key, "TYPE"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( sv_2mortal( (SV*)av ) ) );
+    while(--i >= 0)
+        av_store(av, i, newSViv(imp_sth->fbh[i].dbtype));
+  }
+  else if (kl==9 && strEQ(key, "PRECISION"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( sv_2mortal( (SV*)av ) ) );
+    while(--i >= 0)
+        av_store(av, i, newSViv(imp_sth->fbh[i].prec));
+  }
+  else if (kl==5 && strEQ(key, "SCALE"))
+  {
+    av = newAV();
+    retsv = sv_2mortal( newRV( sv_2mortal( (SV*)av ) ) );
+    while(--i >= 0)
+        av_store(av, i, newSViv(imp_sth->fbh[i].scale));
+  }
+  else if( ( Attribute = getStatementAttr( key, kl ) ) >= 0 )
+  {
+    SQLRETURN ret;
+    char buffer[128]; /* should be big enough for any attribute value */
+    SQLPOINTER ValuePtr = (SQLPOINTER)buffer;
+    SQLINTEGER BufferLength = sizeof( buffer );
+    SQLINTEGER StringLength;
 
-  } else if (kl==13 && strEQ(key, "NUM_OF_PARAMS")) {
-      HV *bn = imp_sth->bind_names;
-      retsv = newSViv( (bn) ? HvKEYS(bn) : 0 );
+    ret = SQLGetStmtAttr( imp_sth->phstmt,
+                          Attribute,
+                          ValuePtr,
+                          BufferLength,
+                          &StringLength );
+    if( SQL_SUCCESS_WITH_INFO == ret &&
+        (StringLength + 1) > BufferLength )
+    {
+      /* local buffer isn't big enough, allocate one */
+      BufferLength = StringLength + 1;
+      Newc( 1, ValuePtr, BufferLength, char, SQLPOINTER );
+      ret = SQLGetStmtAttr( imp_sth->phstmt,
+                            Attribute,
+                            ValuePtr,
+                            BufferLength,
+                            &StringLength );
+    }
 
-  } else if (kl==4 && strEQ(key, "NAME")) {
-      av = newAV();
-      retsv = newRV((SV*)av);
-      while(--i >= 0)
-          av_store(av, i, newSVpv((char *)imp_sth->fbh[i].cbuf,0));
+    ret = check_error( sth, ret, "Error retrieving statement attribute" );
+    if( SQL_SUCCESS == ret )
+    {
+      switch( Attribute )
+      {
+        /* Booleans */
+        case SQL_ATTR_CURSOR_HOLD:
+        case SQL_ATTR_DEFERRED_PREPARE:
+        case SQL_ATTR_EARLYCLOSE:
+        case SQL_ATTR_NOSCAN:
+        case SQL_ATTR_PREFETCH:
+        case SQL_ATTR_RETRIEVE_DATA:
+          if( *(SQLINTEGER*)ValuePtr )
+            retsv = &sv_yes;
+          else
+            retsv = &sv_no;
+          break;
 
-  } else if (kl==8 && strEQ(key, "NULLABLE")) {
-      av = newAV();
-      retsv = newRV(sv_2mortal((SV*)av));
-      while(--i >= 0)
-          av_store(av, i,
-                   (imp_sth->fbh[i].nullok == 1) ? &sv_yes : &sv_no);
+        /* Integers */
+        case SQL_ATTR_CONCURRENCY:
+        case SQL_ATTR_MAX_LENGTH:
+        case SQL_ATTR_MAX_ROWS:
+        case SQL_ATTR_OPTIMIZE_FOR_NROWS:
+        case SQL_ATTR_QUERY_OPTIMIZATION_LEVEL:
+        case SQL_ATTR_QUERY_TIMEOUT:
+        case SQL_ATTR_ROW_NUMBER:
+        case SQL_ATTR_TXN_ISOLATION:
+          retsv = sv_2mortal( newSViv( (IV)( *(SQLINTEGER*)ValuePtr ) ) );
+          break;
 
-  } else if (kl==10 && strEQ(key, "CursorName")) {
-      char cursor_name[256];
-      SQLSMALLINT cursor_name_len;
-      SQLRETURN ret =
-            SQLGetCursorName(imp_sth->phstmt, (SQLCHAR *)cursor_name,
-                             sizeof(cursor_name), &cursor_name_len);
-      ret = check_error( sth, ret, "SQLGetCursorName failed" );
-      if (ret < 0)
-          return Nullsv;
-      else
-          retsv = newSVpv(cursor_name, cursor_name_len);
+        default:
+          break;
+      }
+    }
 
-  } else if (kl==4 && strEQ(key, "TYPE")) {
-      av = newAV();
-      retsv = newRV(sv_2mortal((SV*)av));
-      while(--i >= 0)
-          av_store(av, i, newSViv(imp_sth->fbh[i].dbtype));
+    if( ValuePtr != (SQLPOINTER)buffer )
+      Safefree( ValuePtr );  /* Free dynamically allocated buffer */
 
-  } else if (kl==9 && strEQ(key, "PRECISION")) {
-      av = newAV();
-      retsv = newRV(sv_2mortal((SV*)av));
-      while(--i >= 0)
-          av_store(av, i, newSViv(imp_sth->fbh[i].prec));
-
-  } else if (kl==5 && strEQ(key, "SCALE")) {
-      av = newAV();
-      retsv = newRV(sv_2mortal((SV*)av));
-      while(--i >= 0)
-          av_store(av, i, newSViv(imp_sth->fbh[i].scale));
-
-  } else {
-      return Nullsv;
+    cacheit = 0; /* Don't cache CLI attributes */
+  }
+  else
+  {
+    return Nullsv;
   }
 
+  if( cacheit )
   { /* cache for next time (via DBI quick_FETCH)    */
     SV **svp = hv_fetch((HV*)SvRV(sth), key, kl, 1);
     sv_free(*svp);
     *svp = retsv;
     (void)SvREFCNT_inc(retsv);    /* so sv_2mortal won't free it    */
   }
-  return sv_2mortal(retsv);
+
+  return retsv;
 }
