@@ -1,5 +1,5 @@
 /*
-   engn/perldb2/dbdimp.h, engn_perldb2, db2_v81, 1.6 00/09/06 15:58:14
+   engn/perldb2/dbdimp.h, engn_perldb2, db2_v81, 1.7 01/03/29 17:06:49
 
    Copyright (c) 1995,1996,1997,1998,1999,2000 International Business Machines Corp.
 */
@@ -22,6 +22,7 @@ struct imp_dbh_st {
   dbih_dbc_t com;                     /* MUST be first element in structure   */
   SQLHENV henv;
   SQLHDBC hdbc;
+  char sqlerrp[9];
 };
 
 
@@ -90,6 +91,7 @@ SQLCHAR sql_state[6];
 #define dbd_init            db2_init
 #define dbd_db_login        db2_db_login
 #define dbd_db_do           db2_db_do
+#define dbd_db_ping         db2_db_ping
 #define dbd_db_commit       db2_db_commit
 #define dbd_db_rollback     db2_db_rollback
 #define dbd_db_disconnect   db2_db_disconnect
